// cmtclr.e
//
// Sat, 04/02/1994 02:54:45
//
// Purpose:
//
//		This extension highlights code comments, constants and
//		string/character literals within C and Assembly sources.
//		Keyword highlighting is supported within C sources. Due to
//		limitations of region handling within epsilon, it is
//		impractical to do this highlighting during code edits (or
//		rather it must be done infrequently for performance reasons).
//
//		The intent of this code is for browsing sources.  During
//		development, code flipping color due to open comments,
//		changes of indent etc. would be annoying.  The standard
//		center-window command (bound to ^L) has been subverted for use
//		as a buffer re-highlight (as well as center-window) command.
//
// Usage:
//
//		Load this module and save your state file.
//
//		There are eight user variables broken into two groups and four
//		new color classes.  These twelve configurable items control
//		behavior for each of the four types of highlighting
//		(comments, constants, literals and keywords).  The first
//		group of variables controls whether highlighting is attempted
//		for that highlight type.  The second group of variables
//		controls the size of the file beyond which highlighting will
//		not be done (for performance reasons) when the buffer is
//		first loaded.  The color classes are used for determining the
//		color which the highlight type is highlighted in.  In
//		summary, for each highlight type there are three variables
//		affecting behavior:
//			highlight_XXXXXs			TRUE if highlight XXXXXs
//			highlight_XXXXX_threshold	do not highlight XXXXX above size
//			highlight_XXXXX_color		color for highlight XXXX
//
//		Thresholds are ignored when re-highlighting (^L).
//
// The modifications to this code are dedicated to the public domain
// with the caveat that Lugaru is welcome to use this within their
// distribution source code which is supplied with Epsilon.  Use it
// any way you want for whatever purposes.
//
//
//	John Kercheval
//	15563 10th Ave NE
//	Seattle, WA 98155-6209
//	INTERNET: johnk@wrq.com
//	Compu$erve: 72450,3702
//
// Future Directions:
//
//		Modification of the primary highlighting engine to support
//		partial buffer highlighting (informed) so that background
//		highlighting could be accomplished (perhaps through the
//		when_idle() hook).  The problem is that we are not guaranteed
//		to be able to run background tasks (ie. when_idle() is not
//		guaranteed to be called).  A dramatic speed improvement could
//		be had (visually anyway) by highlighting only the visible
//		portion of the buffer (from the top of the file), then a
//		background highlight could fill the rest.  The problem with
//		this approach is that buffer viewing commands would need to
//		be hooked to make sure that the appropriate parts of the
//		buffer were always highlighted.  This is a reasonably large
//		job which may require a fair amount of future maintainence
//		and would not necessarily gain much advantage.
//
//		Currently the ASM highlighting does not highlight the COMMENT
//		block supported by MASM/TASM syntax.  This syntactic element
//		is rarely used in practice and is not of high priority.
//
//		ASM keyword highlighting is less useful then keyword
//		highlighting within C and is thus not currently implemented.
//		This could be easily changed if there is need.
//
// Modification History:
//
//		Tue, 08/25/1992 11:51:33 V1.0 initial release.
//		Wed, 07/14/1993 22:28:07 V1.1 Update for V6.5
//		Mon, 11/15/1993 22:47:24 V1.2 Fix spot allocation bug.
//			Thanks to Gary Hill for pointing out the bug and
//			supplying a reasonable algorithm for its resolution.
//			The overhead required adds up to a 2X performance hit on
//			initial highlight (session restore) but is quite a bit
//			faster for multiple re-highlights.
//		Sat, 11/20/1993 00:40:58 V1.3 Bow to pressure and add
//			highlighting for string and character literals.
//			Add BOOLEAN control variables for syntax elements.
//			The literals add another 5%-10% performance hit
//			(depending upon coding standards).
//		Fri, 04/01/1994 21:29:57 V1.4 Add keyword and constant
//			highlighting.  This is the daddy of all performance hits
//			and can result in a 2X-3X reduction in performance.  Not
//			for the faint of heart.  The key bottleneck is search
//			speeds for this one with the extra highlighting running a
//			close second.
//		Sat, 04/02/1994 11:23:19 V1.5 Add threshold values for
//			limiting highlighting on larger files.  This is
//			reasonable trade-off of highlighting vs. performance.
//		Mon, 04/04/1994 19:15:11 V1.6 Correct a bug highlighting
//			keywords at the end of the buffer.  Correct a bug in the
//			binary search algorithm for keywords.
//		Tue, 04/05/1994 19:22:14 V1.7 Clean up some debugging output. 
//		Tue, 05/31/1994 20:02:26 V1.8 Correctly handle Hex literals.\
//			Use RE_FASTEST flag for searching (3%-5% gain).  Reduce
//			the spot chunk size to reduce memory requirements editing
//			large numbers of files (which tend to need fewer spots
//			than the previous value).  This allows for finer
//			granularity, but results in slightly larger numbers of
//			allocations during initial highlighting.  
//		Fri, 06/03/1994 20:04:39 V1.9 Reuse the spotlist on
//			re-highlight to avoid memory fragmentation and spot
//			allocation overhead.  This also resulted in a significant
//			performance improvement on re-highlighting (20%-30%
//			including changes by creating macros out of well used
//			procedures and some hand tuning).  Total performance
//			enhancement over V1.7 (re-highlighting) is 40% (highlight
//			of the entire buffer using spot administration and all
//			highlight elements in 60% of the time it took V1.7).
//		Fri, 06/03/1994 23:25:48 V2.0 Add a hook to check_drop_buffer 
//			to clean up allocated memory when the buffer goes away.
//			Correct a problem when other C and ASM mode hooks are
//			installed (need to call the replaced function).  Add a
//			hook to switch_to_buffer so that only those buffers which
//			are going to attach to a window will be highlighted
//			(dramatically speeding up initial startup if many files
//			are in the session file and reducing initial memory
//			requirements).  Looking at V1.1 in comparison to V2.0,
//			simple comment and string highlighting is 5%-10% faster
//			for the first highlight and a whopping 50% faster for
//			re-highlights (with spot administration enabled).  With
//			full keyword and constant highlighting enabled
//			highlighting a buffer takes 50% longer than without those
//			elements enabled.  The end result is a recovery of the 2X
//			loss introducing spot administration in V1.2 and have
//			dramatically limited the performance penalty exacted for
//			constant and keyword highlighting introduced in V1.4.
//		Wed, 06/08/1994 20:32:22 V2.1 Use try_calling to call the C
//			and ASM mode hooks since these hooks may not actually
//			exist. 
//

#include <eel.h>

#define BOOLEAN int
#define TRUE 1
#define FALSE 0


//
// The following variables determine the elements which are
// highlighted.
//
//
// highlight comments?
//
user BOOLEAN highlight_comments = TRUE;

//
// highlight strings and characters in quotes?
//
user BOOLEAN highlight_literals = TRUE;

//
// highlight number constants?
//
user BOOLEAN highlight_constants = TRUE;

//
// highlight keywords? (only supported for C)
//
user BOOLEAN highlight_keywords = TRUE;


//
// The following variables determine the threshold beyond which
// elements are not highlighted on buffer load.
//
user int highlight_comment_threshold = 100000;
user int highlight_literal_threshold = 100000;
user int highlight_constant_threshold = 15000;
user int highlight_keyword_threshold = 10000;


//
// The following color classes are used to determine highlight colors.
//
color_class highlight_comment_color = {
	COLOR(colGREEN, colBLACK), MCNORM
};
color_class highlight_literal_color = {
	COLOR(colWHITE, colBLACK), MCNORM
};
color_class highlight_constant_color = {
	COLOR(colWHITE, colBLACK), MCNORM
};
color_class highlight_keyword_color = {
	COLOR(colBROWN, colBLACK), MCNORM
};


//
// maintain a spot list? (speeds up re-highlighting dramatically but
// may slow down initial highlighting a small amount).  Highly
// recommended that this is set TRUE for memory management.
//
user BOOLEAN highlight_spot_administration = TRUE;


//
// define the search patterns for keywords and number constants.
// keywords are preprocessor directives and any element listed in the
// variable below.  Constants are true numbers and hex values (U and
// L are allowed suffixes).
//
#define NUMBER_PATTERN	"0?[xX?]?[0-9a-fA-F]+[uU]?[lL]?"
#define KEYWORD_PATTERN "(%#[ \t]*)?_?_?[a-zA-Z]+"
#define NUMBER_AND_KEYWORD_PATTERN KEYWORD_PATTERN "|" NUMBER_PATTERN


//
// C keyword list, maintained in sorted order
//
#define MAX_KEYWORD_SIZE 10
char highlight_C_keyword_list[][MAX_KEYWORD_SIZE+1] = 
{
	"_asm",
	"_based",
	"_cdecl",
	"_emit",
	"_except",
	"_export",
	"_far",
	"_fastcall",
	"_finally",
	"_fortran",
	"_huge",
	"_interrupt",
	"_loadds",
	"_near",
	"_pascal",
	"_saveregs",
	"_segment",
	"_segname",
	"_self",
	"_stdcall",
	"_syscall",
	"_try",
	"asm",
	"auto",
	"break",
	"case",
	"catch",
	"char",
	"class",
	"const",
	"continue",
	"default",
	"delete",
	"do",
	"double",
	"else",
	"enum",
	"extern",
	"float",
	"for",
	"friend",
	"goto",
	"if",
	"inline",
	"int",
	"long",
	"new",
	"operator",
	"private",
	"protected",
	"public",
	"register",
	"return",
	"short",
	"signed",
	"sizeof",
	"static",
	"struct",
	"switch",
	"template",
	"this",
	"throw",
	"try",
	"typedef",
	"union",
	"unsigned",
	"virtual",
	"void",
	"volatile",
	"while"
};


//
// This is the region handle
//
buffer int highlight_region_handle = 0;
buffer int highlight_region_count = 0;


//
// These variables control the allocated spot array
//
#define SPOTLIST_CHUNK_SIZE 100
buffer int spotcount = 0;
buffer int spotalloc = 0;
buffer int spotavail = 0;
buffer spot *spotlist = 0;


//
// These variables control initial highlighting
//
#define HIGHLIGHT_TYPE_UNKNOWN 0
#define HIGHLIGHT_TYPE_C 1
#define HIGHLIGHT_TYPE_ASM 2
buffer BOOLEAN highlight_buffer_initialized = FALSE;
buffer int highlight_buffer_type = HIGHLIGHT_TYPE_UNKNOWN;
buffer int highlight_buffer_highlight_count = 0;


//
// These variables are here only as external declarations
//
keytable c_tab;
keytable asm_tab;


////////////////////////////////////////////////////////////////////////////
//
// highlight_check_threshold will set all highlight variables based on the
// current threshold level
//
void
highlight_check_threshold()
{
	int buflength = size();
	
	if (highlight_comment_threshold &&
		highlight_comment_threshold < buflength) {
		highlight_comments = FALSE;
	}

	if (highlight_literal_threshold &&
		highlight_literal_threshold < buflength) {
		highlight_literals = FALSE;
	}

	if (highlight_constant_threshold &&
		highlight_constant_threshold < buflength) {
		highlight_constants = FALSE;
	}

	if (highlight_keyword_threshold &&
		highlight_keyword_threshold < buflength) {
		highlight_keywords = FALSE;
	}
}


////////////////////////////////////////////////////////////////////////////
//
// C and ASM mode hooks will highlight a buffer which is already
// viewable but has an unknown highlight type.
//
new_highlight_c_mode_hook()
{
	if (highlight_buffer_initialized &&
		highlight_buffer_type == HIGHLIGHT_TYPE_UNKNOWN) {
	
		save_var highlight_comments;
		save_var highlight_literals;
		save_var highlight_constants;
		save_var highlight_keywords;

		highlight_check_threshold();

		highlight_c_syntax();
		restore_vars();
	}
	highlight_buffer_type = HIGHLIGHT_TYPE_C;
	try_calling("highlight_c_mode_hook");
}
REPLACE_FUNC("highlight", "c_mode_hook")

	
new_highlight_asm_mode_hook()
{
	if (highlight_buffer_initialized &&
		highlight_buffer_type == HIGHLIGHT_TYPE_UNKNOWN) {
	
		save_var highlight_comments;
		save_var highlight_literals;
		save_var highlight_constants;
		save_var highlight_keywords;

		highlight_check_threshold();

		highlight_asm_syntax();
		restore_vars();
	}
	highlight_buffer_type = HIGHLIGHT_TYPE_ASM;
	try_calling("highlight_asm_mode_hook");
}
REPLACE_FUNC("highlight", "asm_mode_hook")

	
////////////////////////////////////////////////////////////////////////////
//
// new_highlight_switch_to_buffer is responsible for initially
// highlighting a buffer before it is viewable.
//
new_highlight_switch_to_buffer(num)
	int num;
{
	save_var bufnum = num;
	
	if (!highlight_buffer_initialized) {
	
		save_var highlight_comments;
		save_var highlight_literals;
		save_var highlight_constants;
		save_var highlight_keywords;

		highlight_check_threshold();

		switch (highlight_buffer_type) {
		case HIGHLIGHT_TYPE_C:
			highlight_c_syntax();
			break;
			
		case HIGHLIGHT_TYPE_ASM:
			highlight_asm_syntax();
			break;
			
		case HIGHLIGHT_TYPE_UNKNOWN:
		default:
			break;
		}
		highlight_buffer_initialized = TRUE;
	}
	restore_vars();
	highlight_switch_to_buffer(num);
}
REPLACE_FUNC("highlight", "switch_to_buffer")

	
////////////////////////////////////////////////////////////////////////////
//
// new_highlight_check_drop_buffer is responsible for freeing allocated
// memory for the buffer spots and spot list before the buffer goes
// away. 
//
new_highlight_check_drop_buffer(bname)
	char *bname;
{
	save_var bufname = bname;
	
	//
	// free any memory associated with the spotlist
	//
	highlight_free_region_spot_list();
	
	restore_vars();
	return highlight_check_drop_buffer(bname);
}
REPLACE_FUNC("highlight", "check_drop_buffer")


////////////////////////////////////////////////////////////////////////////
//
// new_highlight_restore_session_hook highlights buffers which are
// visible after the session restore completes.
//
new_highlight_restore_session_hook(b)
{
	int index;
	int previous_bufnum;
	save_var bufnum = b;
	save_var window_number;
	
	
	for (index = 0; index < number_of_windows(); index++) {
		window_number = index;
		previous_bufnum = bufnum;
		bufnum = window_bufnum;
		
		if (!highlight_buffer_initialized) {

			save_var highlight_comments;
			save_var highlight_literals;
			save_var highlight_constants;
			save_var highlight_keywords;

			highlight_check_threshold();

			switch (highlight_buffer_type) {
			case HIGHLIGHT_TYPE_C:
				highlight_c_syntax();
				break;

			case HIGHLIGHT_TYPE_ASM:
				highlight_asm_syntax();
				break;

			case HIGHLIGHT_TYPE_UNKNOWN:
			default:
				break;
			}
			highlight_buffer_initialized = TRUE;
		}
		bufnum = previous_bufnum;
	}
	
	//
	// call the previous hook
	//
	restore_vars();
	highlight_restore_session_hook(b);
}

REPLACE_FUNC("highlight", "restore_session_hook")

	
////////////////////////////////////////////////////////////////////////////
//
// C and ASM highlight commands will center window (as does a
// standard ^L) and then rehighlight the buffer according to buffer
// mode.
//
command
highlight_c_buffer()
	on c_tab[CTRL('L')]
{
	center_window();
	note("Highlighting buffer.");
	highlight_c_syntax();
	note("Highlighting complete (%d items).", highlight_region_count);
}

command
highlight_asm_buffer()
	on asm_tab[CTRL('L')]
{
	center_window();
	note("Highlighting buffer.");
	highlight_asm_syntax();
	note("Highlighting complete (%d items).", highlight_region_count);
}


////////////////////////////////////////////////////////////////////////////
//
// determine membership in the keyword list
//
BOOLEAN
highlight_is_c_keyword(start, end)
	int start;
	int end;
{
	char string[MAX_KEYWORD_SIZE+1];
	char *c_current = string;
	char c = character(start);

	int keyword_index;
	int start_index;
	int low_index;
	int high_index;
	
	int max_keyword_index = (sizeof(highlight_C_keyword_list) / 
		sizeof(highlight_C_keyword_list[0])) - 1;
	
	//
	// do some initial filtering based on the first char
	//
	if (end - start > MAX_KEYWORD_SIZE) {
		//
		// there is no keyword this large
		//
		return FALSE;
	}
	if (index("hjklmoqxyz", c)) {
		//
		// starts with a letter which does NOT start a keyword
		//
		return FALSE;
	}
	if (c == '#') {
		//
		// all preprocessor directives are keywords
		//
		return TRUE;
	}

	//
	// obtain the rest of the keyword
	//
	grab(start, end, string);

	if (c == '_') {
		if (*(c_current+1) == '_') {
			//
			// one or two underscores are allowed on these keywords
			//
			c_current++;
		}
	}

	//
	// perform a binary search to find a keyword with the first
	// character the same as the first letter.
	//
	low_index = 0;
	high_index = max_keyword_index;
	start_index = (low_index + high_index) / 2;
	while (low_index < high_index) {
		if (c > highlight_C_keyword_list[start_index][0]) {
 			low_index = start_index + 1;
		} else if (c < highlight_C_keyword_list[start_index][0]) {
			high_index = start_index - 1;
		} else {
			high_index = low_index = start_index;
		}
		start_index = (low_index + high_index) / 2;
	}
			
	//
	// move through the index list until we have determined
	// membership.  First go forward to the end of the section, then
	// go back from the start index if still not found.
	//
	for (keyword_index = start_index;
		 keyword_index <= max_keyword_index;
		 keyword_index++) {
		if (!strcmp(c_current, highlight_C_keyword_list[keyword_index])) {
			//
			// successful match
			//
			return TRUE;
		}
		if (c != highlight_C_keyword_list[keyword_index][0]) {
			//
			// we are past the elements with this first letter
			//
			break;
		}
    }
	if (start_index) {
		for (keyword_index = start_index-1;
			 keyword_index >= 0;
			 keyword_index--) {
			if (!strcmp(c_current, highlight_C_keyword_list[keyword_index])) {
				//
				// successful match
				//
				return TRUE;
			}
			if (c != highlight_C_keyword_list[keyword_index][0]) {
				//
				// we are past the elements with this first letter
				//
				break;
			}
		}
	}
	
	//
	// no match found
	//
	return FALSE;
}


////////////////////////////////////////////////////////////////////////////
//
// highlight_free_region_spot_list frees the spot list and all associated spots
//
void
highlight_free_region_spot_list()
{
	if (highlight_spot_administration && 
		spotlist) {
		spotalloc = 0;
		free(spotlist);
		spotlist = 0;
		spotavail = 0;
	}
}


////////////////////////////////////////////////////////////////////////////
//
// alloc_region_spot() will obtain a spot and place it into the spot
// list.  First allocate the spotlist, or enlarge it if required,
// then add the spot to the spotlist,  if we already have an
// allocated spot then reuse it, otherwise get the new spot.
//
#define alloc_region_spot(return_spot) \
{ \
	if (highlight_spot_administration) { \
		if (spotavail <= spotcount) { \
			spotavail += SPOTLIST_CHUNK_SIZE; \
			if (spotlist) { \
				spotlist = (spot*) realloc(spotlist, spotavail * sizeof(spot)); \
			} else { \
				spotlist = (spot*) malloc(spotavail * sizeof(spot)); \
			} \
		} \
	 	if (spotcount < spotalloc) { \
			return_spot = spotlist[spotcount]; \
			*return_spot = point; \
		} else { \
			return_spot = alloc_spot(); \
			spotalloc++; \
			spotlist[spotcount] = return_spot; \
		} \
		spotcount++; \
	} else { \
		return_spot = alloc_spot(); \
	} \
}


////////////////////////////////////////////////////////////////////////////
//
// highlight_init_region_handles assigns region handles for syntax
// elements.  It first removes old regions and inits the current spot
// list then creates the initial handle (the color used for the
// initial region is not important).  This macro assumes the
// variables spot start_spot and spot end_spot. 
//
#define highlight_init_region_handles() \
{ \
	if (highlight_region_handle) { \
		remove_region(highlight_region_handle); \
		highlight_region_handle = 0; \
	} \
	spotcount = 0; \
	highlight_region_count = 0; \
	alloc_region_spot(start_spot); \
	alloc_region_spot(end_spot); \
	highlight_region_handle = add_region(start_spot, end_spot, \
										 color_class highlight_comment_color, \
										 REGNORM, 0); \
}


////////////////////////////////////////////////////////////////////////////
//
// highlight_region will highlight all of the text between start and
// point by allocating a couple of spots and adding a region.
// Depends on the existence of a variable int cur_point, spot
// start_spot and spot end_spot for performance.
//
#define highlight_region(region_color, start) \
{ \
	cur_point = point; \
	alloc_region_spot(end_spot); \
	point = start; \
	alloc_region_spot(start_spot); \
	add_region(start_spot, end_spot, region_color, \
			   REGNORM, highlight_region_handle); \
	highlight_region_count++; \
	point = cur_point; \
} 


////////////////////////////////////////////////////////////////////////////
//
// highlight_string will highlight all of the text between the
// narrowed region start characters from the beginning of the buffer
// and end characters from the end of the buffer.  This macro assumes
// acces to a variable int buflength.
//
#define highlight_string(search_string, start, end) \
{ \
  	int previous_point = point; \
	narrow_end = end; \
	point = start; \
	while (re_search(RE_FORWARD, search_string)) { \
		if (!is_word_char(matchstart - 1) && \
			!is_word_char(point)) { \
			if (highlight_keywords && \
				highlight_is_c_keyword(matchstart, point)) { \
				highlight_region(color_class highlight_keyword_color, \
								 matchstart); \
			} else if (highlight_constants && \
					   isdigit(character(matchstart))) { \
				highlight_region(color_class highlight_constant_color, \
								 matchstart); \
			} \
		} \
	} \
	point = previous_point; \
	narrow_end = 0; \
}


////////////////////////////////////////////////////////////////////////////
//
// highlight_c_syntax() will highlight all C elements to be
// hightlighted in the current (C syntax) buffer.
//
void
highlight_c_syntax()
{
	int buflength = size();
	int start = 0;
	int cur_point;

	spot start_spot;
	spot end_spot; 
	
	char *search_string;

	BOOLEAN isComment;
    char c;

	save_spot point = 0;
	save_spot mark = 0;

	save_var case_fold = 0;
	save_var regex_first_end = 1;
	save_var regex_shortest = 1;
	save_var narrow_start; 
	save_var narrow_end; 

	highlight_buffer_highlight_count++;
	
	//
	// Highlighting only makes sense in color or gray scale
	//
	if (monochrome) {
		return;
	}

	//
	// Do not waste time if we are not highlighting any syntax
	// elements.
	//
	if (!highlight_comments &&
		!highlight_literals &&
		!highlight_constants &&
		!highlight_keywords) {
		return;
	}

	//
	// Place the first (empty) region to obtain a region handle
	//
	highlight_init_region_handles();

	//
	// initialize the search string
	//
	if (highlight_constants &&
		highlight_keywords) {
		search_string = NUMBER_AND_KEYWORD_PATTERN;
	} else {
		if (highlight_constants) {
 			search_string = NUMBER_PATTERN;
		}
		if (highlight_keywords) {
 			search_string = KEYWORD_PATTERN;
		}
	}

	//
	// Parse through the buffer
	//
    while (point != buflength) {
		if (!re_search(RE_FORWARD | RE_FASTEST, "[/\"\']")) {
			//
			// do keyword and constant highlighting at end of buffer
			//
			point = buflength;
			if (point > mark+1 &&
				(highlight_constants ||
				 highlight_keywords)) {
				highlight_string(search_string, mark, 0);
			}
		} else {
			//
			// obtain the match begin and the character
			//
			start = matchstart;
			c = character(start);

			//
			// do keyword and constant highlighting between strings
			// and comments
			//
			if (point > mark+1 &&
				(highlight_constants ||
				 highlight_keywords)) {
				highlight_string(search_string, mark, buflength - start - 1);
			}

			//
			// do comment and literal highlighting
			//
			switch (c) {
				case '/':
					switch (curchar()) {
						case '/':
							if (!nl_forward()) {
								point = buflength;
							}
							isComment = TRUE;
							break;
						case '*':
							if (!search(1, "*/")) {
								point = buflength;
							}
							isComment = TRUE;
							break;
						default:
							isComment = FALSE;
							break;
					}
					if (highlight_comments &&
						isComment) {
						highlight_region(color_class highlight_comment_color,
										 start);
					}
					mark = point;
					break;

				case '\"':
				case '\'':
					while (curchar() != c &&
						   point < buflength) {
						if (curchar() == '\\') {
							point++;
						}
						point++;
					}
					point++;
					if (highlight_literals) {
						highlight_region(color_class highlight_literal_color,
										 start);
					}
					mark = point;
					break;

				default:
					break;
			}
		}
    }
}


////////////////////////////////////////////////////////////////////////////
//
// highlight_asm_syntax() will highlight all elements to be
// highlighted within the current (ASM syntax) buffer.
//
//
void
highlight_asm_syntax()
{
	int buflength = size();
	int start = 0;
	int cur_point;

	spot start_spot;
	spot end_spot; 
	
	char c;

	save_spot point = 0;
	save_spot mark = 0;
	save_var narrow_start; 
	save_var narrow_end; 

	highlight_buffer_highlight_count++;
		
	//
	// Highlighting only makes sense in color or gray scale
	//
	if (monochrome) {
		return;
	}

	//
	// Do not waste time if we are not highlighting any syntax
	// elements.
	//
	if (!highlight_comments &&
		!highlight_literals &&
		!highlight_constants) {
		return;
	}

	//
	// Place the first (empty) region to obtain a new region handle
	//
	highlight_init_region_handles();

	//
	// Parse through the buffer
	//
    while (point != buflength) {
		if (!re_search(RE_FORWARD | RE_FASTEST, "[;>\"\']")) {
			//
			// do constant highlighting at end of buffer
			//
			point = buflength;
			if (point > mark+1 &&
				highlight_constants) {
				highlight_string(NUMBER_PATTERN, mark, 0);
			}
		} else {
			//
			// obtain the match begin and the character
			//
			start = matchstart;
			c = character(start);

			//
			// do constant highlighting between strings and comments
			//
			if (point > mark+1 &&
				highlight_constants) {
				highlight_string(NUMBER_PATTERN, mark, buflength - start - 1);
			}

			//
			// highlight the comment or literal
			//
			switch (c) {
				case ';':
					if (!nl_forward()) {
						point = buflength;
					}
					if (highlight_comments) {
						highlight_region(color_class highlight_comment_color,
										 start);
					}
					mark = point;
					break;

				case '\"':
				case '\'':
					while (curchar() != c &&
						   point < buflength) {
						point++;
					}
					point++;
					if (highlight_literals) {
						highlight_region(color_class highlight_literal_color,
										 start);
					}
					mark = point;
					break;

				case '<':
					while (curchar() != '>' &&
						   point < buflength) {
						if (curchar() == '!') {
							point++;
						}
						point++;
					}
					point++;
					if (highlight_literals) {
						highlight_region(color_class highlight_literal_color,
										 start);
					}
					mark = point;
					break;

				default:
					break;
			}
		}
	}
}
